;**************************************************************************
;***			PIC Programmer					***
;**************************************************************************
;Communication:
;Command from host:
;<cmd> [<ldat>,<hdat>] <chksum>
;reply from programmer:
;NAK			(case error)
;ACK 			(command completed ok, on some operations)
;<ldat>,<hdat> [<ldat>,<hdat>...]<chksum>	(returned data)
;
;cmd: 	bits 0..6 = command code for PIC, 
;	bits 7:6 = 00: no data I/O,
;	bits 7:6 = 01: return data (read),
;	bits 7:6 = 10: send data (load),
;	bits 7:6 = 11: bits 0..6 = special command to programmer:
;			0xC0: ready sense (does nothing but replying ACK)
;			0xC1: read status
;			0xC2: exit programming mode
;			0xC3: enter programming mode (->ACK)
;			0xD4: set VPP value 
;			0xD5: set repeat count
;			0xE6: wait given time (16 bits, microseconds)
;			0xC7: run application in target PIC
;			0xC8: start into bootloader

#define ACK 0x06
#define NAK 0x15
  
    radix dec		    ;default radix: decimal 

#define LED_PORT	PORTC	
#define LED_PIN		2		
#define PWM_PORT	PORTA		;PWM out = RA5 (PWM = P2A)
#define PWM_PIN		5
#define TRES_PORT	PORTA		;apply target reset  
#define TRES_PIN	4		;(pull MCLR/VPP low by output high)
#define VFB_PORT	PORTA		;programming voltage feedback
#define VFB_PIN		0		;(voltage divider: *0.18)		
#define SWVDD_PORT	PORTC		;VDD switch (output low = ON)
#define SWVDD_PIN	0
#define SWVPP_PORT	PORTC		;VPP switch (output high = ON)
#define SWVPP_PIN	1
#define VDD_PORT	PORTC		;connected to target VDD (through 4k7)
#define VDD_PIN		3
#define RX_PORT		PORTC
#define RX_PIN		5
#define TX_PORT		PORTC
#define TX_PIN		4
#define PDAT_PORT	PORTA
#define PDAT_PIN	2
#define PCLK_PORT	PORTA
#define PCLK_PIN	1

#define BAUDRATE	115200		;COM baudrate 
#define MAXPWMPULSE	12		;max. PWM pulse width (cycles)
#define IVPP_LIMIT	0x50		;max allowed voltage drop by I(VDD)
 
;#define SIMULATOR

FOSC set 32000000			;clock oscillator frequency

#ifdef __16F1825	
  #include <p16F1825.inc>         	; processor specific variable definitions
;  __config _CONFIG1, _FOSC_INTOSC & _WDTE_OFF & _PWRTE_ON & _MCLRE_ON & _CP_OFF & _BOREN_ON & _IESO_OFF & 0x3FFF
;  __config _CONFIG2, _WRT_OFF & _PLLEN_ON & _BORV_HI & _LVP_OFF & 0x3FFF
#endif

 ; ------ Calculation for EUSART baudrate generator: 
BAUDGEN set (FOSC/2)/BAUDRATE 
 if (BAUDGEN & 1)
BAUDGEN set BAUDGEN/2
 else
BAUDGEN set BAUDGEN/2-1
 endif
	CBLOCK	0x20
fifoWr
fifoRd
timReg1: 2
conTimer: 2
conState
conVar: 2

r16_a:	2			;some 16-bit registers for math
r16_b:	2
r16_c:	2
r16_x:	2
vppValue 			;VPP voltage level setting (*0.1V)
command: 3			;command from host
chksum				;checksum byte for command validation
cmdstate			
repeatCnt


	ENDC

	CBLOCK 	0x70		;common RAM: 
temp				;temporary storage
flags				;some status flags
w16:	2			;16 bit WREG
halfDplxCnt			;!=0 if Rx = echo from single wire expected
timeAccu			;used to generate 1ms intervalls
mscnt: 	2
setVP:	2			;VPP voltage setting
ad_res:	2			;AD result
	ENDC
	
;----------- flags bit definitions:
ADLOCK	equ	0		;disable AD conversion by ISR
VPPCTRL equ	1		;wait for VPP ready to activate VPP
TRGCON	equ	3		;detected/assumed target connection
TRGXV	equ	4		;target supplied by external power
VPPRDY	equ	5		;VPP ready to turn on
VPPACTV equ	6		;set if VPP is on
CONERR	equ	7		;detected a wrong connection of target 

fifo 	equ	0x2100		;linear address of 256 bytes FIFO buffer

	errorlevel -302,-306			;avoid lots of warnings about banking

;---------------- Macro: move word (byte order: low,high)
mov16	macro 	source, dest	
	movf	source,W
	movwf	dest	
	movf	source+1,W
	movwf	dest+1
	endm	

;*************************************************************************************
;Program Code:

	ORG 0
	banksel	OSCCON			;after reset, MPU is running on 500 kHz clock
	movlw	b'01110000'		;set clock to 32MHz:
	movwf	OSCCON
	goto	startup

;----------------------------- Interrupt Service Routine
;IRQ source:	Used for:
;	TMR2	triggers ADC, system milliseconds counter
;	ADC	setting PWM (boost converter) according to measured voltage
; EUSART RX	Rx data transfered into Rx-FIFO

	ORG 4
interrupt 
	banksel	PIE1
	movf	PIE1,W			;mask enabled interrupts
	banksel	PIR1
	andwf	PIR1,W			;& flags = true interrupt requests

	btfss	WREG,TMR2IF		;TMR2 interrupt?
	goto	isrSkipT2		;no

;--------------- TMR2 ISR, called every 128 s
	bcf	PIR1,TMR2IF

	banksel	ADCON0
	btfss	flags,ADLOCK		;skip if AD converter is used by main
	bsf	ADCON0,GO		;issues another AD-conversion

	movlw	32
	addwf	timeAccu,F		;add 128s to accu
	btfss	STATUS,C
	retfie

;--------------- executes approx. every 1 ms
	bcf	flags,VPPRDY

	movf	ad_res,W
	subwf	setVP,W			;setVP - ad_res
	movwf	FSR0			;(save LSB)
	movf	ad_res+1,W
	subwfb	setVP+1,W
	btfss	STATUS,Z
	goto	isrSkipNRDY		;MSB != 0: assume not ready
	movlw	5
	subwf	FSR0,W
	btfss	STATUS,C		;skip if difference >5		
	
	bsf	flags,VPPRDY

isrSkipNRDY
	movlw	8
	addwf	timeAccu,F
	incfsz	mscnt,F			;adjust system tick counter	
	retfie
	incf	mscnt+1,F
	retfie

isrSkipT2
	btfss	WREG,ADIF		;AD interrupt?
	goto	isrSkipADC		;no

;--------------- ADC conversion complete interrupt
;	movlw	1<<LED_PIN
;	xorwf	LED_PORT,F

	bcf	PIR1,ADIF
	banksel	ADRESL
	mov16	ADRESL,ad_res		;save actual value
	movf	setVP,W			;ADRES - setVP -> ADRES 
	subwf	ADRESL,F
	movf	setVP+1,W
	subwfb	ADRESH,F

	movlw	0
	btfsc	STATUS,C
	goto	isrSetPWM		;ADRES >= setVP: turn PWM off

	movlw	MAXPWMPULSE
	incfsz	ADRESH,F
	goto	isrSetPWM		; ADRESH < -1: choose maximum pulse width
	
	addwf	ADRESL,F
	btfss	STATUS,C
	goto	isrSetPWM		; ADRESL < -MAXPWMPULSE: choose maximum pulse width

	subwf	ADRESL,W
	sublw	0			; abs()

isrSetPWM
	banksel	CCPR2L
	movwf	CCPR2L
	retfie

isrSkipADC
	btfss	WREG,RCIF		;some data in RX register?
	retfie				;(some other interrupt - should not occur)
	
isrRxByte
	movlw	high fifo
	movwf	FSR0H
	movf	fifoWr,W
	movwf	FSR0L

	decf	halfDplxCnt,F
	incfsz	halfDplxCnt,W
	goto	SkipRxByte		;halfDplxCnt != 0: discard echo of own transmission
	
	clrf	halfDplxCnt
	incf	fifoWr,F
SkipRxByte
	banksel	RCREG
	movf	RCREG,W
	movwf	INDF0
	retfie

;************************* Initialization:
startup
	banksel	0				;select standard bank (0)
	clrf	FSR0H
	movlw	0x20
	movwf	FSR0L
initclr						;clr all vars (memory range 0x20..0x7F)
	clrf	INDF0
	incf	FSR0L,F
	btfss	FSR0L,7
	goto	initclr

;--------------- Setup Timer 0
;	banksel	OPTION_REG
;	movlw	b'11011000'			;Timer 0 clock = 8 MHz
;	movwf	OPTION_REG			
 
;--------------- Setup Timer 1
	banksel	T1CON
	movlw	b'00110000'			;T1 counting @1MHz
	movwf	T1CON
;	clrf	TMR1L
;	clrf	TMR1H

;--------------- Initialize I/O Ports:
	banksel	PORTA
	clrf	PORTA
	clrf	PORTC
	banksel LATA
	clrf	LATA
	clrf	LATC
	bsf	SWVDD_PORT,SWVDD_PIN		;disable VDD output (active low!)

	banksel	TRISA
	clrf	TRISA
	clrf	TRISC
	bsf	RX_PORT,RX_PIN			;RXD = input
	bsf	VFB_PORT,VFB_PIN		;voltage feedback = input

	banksel	ANSELA
	clrf	ANSELA
	clrf	ANSELC
	bsf	VFB_PORT,VFB_PIN		;voltage feedback = analog input

;--------------- Initialize FVR and ADC:
	banksel	FVRCON
	movlw	b'10000010'			;ref. for ADC = 2.048 V 
	movwf	FVRCON
	banksel	ADCON0
	movlw	b'11100011'			;right just., 0.5MHz, PREF=VFR
	movwf	ADCON1
	movlw	1
	movwf	ADCON0				;select input: AN0, enable ADC 

;--------------- Initialize EUSART:
	banksel	TXSTA
	movlw	low BAUDGEN
	movwf	SPBRGL
    if (high BAUDGEN)
	movlw	high BAUDGEN
	movwf	SPBRGH
    else
	clrf	SPBRGH
    endif
	movlw	1<<BRGH	| 1<<TXEN   	;high speed baudrate, enable Tx 
	movwf	TXSTA
	movlw	1<<SPEN | 1<<CREN   	;enable serial Tx & Rx
	movwf	RCSTA
	movlw	1<<BRG16
	movwf	BAUDCON

;--------------- Initialize TMR2 (PWM for boost converter):
	banksel	T2CON
	clrf	TMR2
	movlw	b'00111100'			;TMR2 on, clk= 8MHz, postscaler = 8
	movwf	T2CON
	bcf	PIR1,TMR2IF
	movlw	127
	movwf	PR2				;periode = 16s

;--------------- Initialize PWM:
	banksel	APFCON1
	bsf	APFCON1,CCP2SEL			;PWM -> RA5
	banksel	TRISA
	bsf	PWM_PORT,PWM_PIN		;PWM = input during initialization
	banksel CCP2CON
	clrf	CCPTMRS				;use TMR2
	clrf	CCPR2L				;set duty = 0%
	movlw	b'00001100'			;PWM output = P2A
	movwf	CCP2CON

	banksel	PIR1
	btfss	PIR1,TMR2IF			;wait for full cycle
	goto 	$-1
	clrf	PIR1				;clr peripheral interrupt flags

	banksel	TRISA
	bcf	PWM_PORT,PWM_PIN		;PWM = output

	banksel	PIE1
	movlw	(1<<TMR2IE) | (1<<ADIE) | (1<<RCIE)
	movwf	PIE1		

	banksel	0

	bsf	INTCON,PEIE
	bsf	INTCON,GIE
	movlw	85				;VPP setting = 8.5V
	movwf	vppValue 
	

;************************* Main Loop:
mainLoop
	banksel	0
	bcf	LED_PORT,LED_PIN	

	btfss	flags,CONERR
	goto	mainDispVpp			;actually no connection error

	btfss	mscnt,7				;fast flashing LED signal misconnected target
	bsf	LED_PORT,LED_PIN	
	goto	mainChkCon		

mainDispVpp
	btfss	flags,TRGCON
	goto	mainChkCon			;VPP-LED only if connected!

	btfsc	SWVPP_PORT,SWVPP_PIN
	bsf	LED_PORT,LED_PIN		;LED = status of VPP on/off

mainChkCon
	call	checkConnection

	btfsc	flags,VPPCTRL			
	call	PrgModeWhenReady		;entering promode as soon as VPP ready...

	movlw	high fifo
	movwf	FSR1H
	movf	fifoRd,W
	movwf	FSR1L
	subwf	fifoWr,W
	btfss	STATUS,Z		;any rx data in buffer?
	call	processRxData		;y: process it!

	goto	mainLoop

;************************* Subroutines:

processRxData
	incf	fifoRd,F
	movf	INDF1,W
	addwf	chksum,F

	movf	cmdstate,W
	brw
	goto	cmdS0			;0: get first byte = cmd code
	goto	cmdS1			;1: first payload data byte or chksum->execute 
	goto	cmdS2			;2: 2nd payload data byte
	goto	cmdS3			;3: chksum->execute 
	goto	cmdS4			;4: receive data byte (-> command+1)
	goto	cmdS5			;5: receive data byte (-> command+2)
	goto	cmdS6			;6: chksum, special command processing

cmdError
	call	sendNAK
	clrf	repeatCnt

cmdComplete
	clrf	chksum
	clrf	cmdstate
	return

;----- state 0: get command code from input buffer
cmdS0
	movf	repeatCnt,F		;repeat not set?
	btfsc	STATUS,Z
	incf	repeatCnt,F		;set loop counter to 1

	movf	INDF1,W			;store command code
	movwf	command
	
	andlw	0xC0
	xorlw	0xC0
	btfsc	STATUS,Z
	goto	specialCmd		;command for programmer device

	btfsc	command,7
	goto	incCmdState		;"Load": next two bytes = payload data

	movlw	3			;set state=3: checksum, command processing
	movwf	cmdstate
	return

;----- state 1: first byte of payload data for "Load"-cmd 
cmdS1
	clrf	repeatCnt		;clear any repeats

;----- state 4: first byte of payload data for special command
cmdS4
	movf	INDF1,W			;first byte after command code
	movwf	command+1 
	goto	incCmdState

;----- state 2 or 5: second byte of payload data for cmd 
cmdS2
cmdS5
	movf	INDF1,W			;2nd byte after command code
	movwf	command+2 

incCmdState
	incf	cmdstate,F
	return

;----- state 3: validate checksum, process command
cmdS3	
	movf	chksum,F
	btfss	STATUS,Z		
	goto	cmdError		;checksum error!

	btfsc	command,6		;read command?
	goto	cmdRdLoop		;y

	btfsc	command,7		;load command?
	goto	cmdLd

singleByteCmd
	call	sendCommand		;send command to PIC
	decfsz	repeatCnt,F		;repeat?
	goto	singleByteCmd
	goto	cmdComplete

;------ Load command: send command code, then 2 bytes of data 
cmdLd
	call	sendCommand		;send command to PIC
	mov16	command+1,w16		;w16 = payload data
	call	sendICSPDAT
	goto	cmdComplete

;------ execute Read command:
cmdRdLoop
	call	sendCommand
	call	getICSPDAT		;read data from PIC -> w16
	movf	w16,W
	subwf	chksum,F
	call	txByte
	movf	w16+1,W
	subwf	chksum,F
	call	txByte
	decfsz	repeatCnt,W
	call	incAddr			;increment address on multi-word read
	decfsz	repeatCnt,F
	goto	cmdRdLoop

	movf	chksum,W
	call	txByte
	goto	cmdComplete
	


;--------- >= 0xC0: special programmer command:
specialCmd
	clrf	repeatCnt
	swapf	command,W		;bit 5:4 = #of data bytes -> bits 1:0
	andlw	3
	sublw	6			;cmdS6 = chksum, special command processing 
	movwf	cmdstate
	return				;next byte = checksum

;-------- state 6: process special command
cmdS6
	clrf	cmdstate		;no further states

	movf	chksum,F
	btfss	STATUS,Z		
	goto	cmdError		;checksum error!

	movf	command,W
	andlw	0x0F			;mask command code
	brw
	goto	sendACK			;cmd 0: ready sense
	goto	readStatus		;cmd 1: read Status flags
	goto	exitPrgMode		;cmd 2: exit programming mode
	goto	enterPrgMode		;cmd 3: enter programming mode
	goto	setVppValue		;cmd 4: set VPP value
	goto	setRptCnt		;cmd 5: set repeat count
	goto	delay_us		;cmd 6: delay by x s
	goto	cmdRun			;cmd 7: run application
	goto	cmdBootldr		;cmd 8: start bootloader
	goto	cmdError		;cmd 9: n.a.
	goto	cmdError		;cmd A: n.a.
	goto	cmdError		;cmd B: n.a.
	goto	cmdError		;cmd C: n.a.
	goto	cmdError		;cmd D: n.a.
	goto	cmdError		;cmd E: n.a.
	goto	cmdError		;cmd F: n.a.



readStatus
 	movf	flags,W			;return status flags
	goto	txByte

setRptCnt
	movf	command+2,W		;repeat next command x times
	movwf	repeatCnt
	return
	
setVppValue
	movf	command+2,W		;new VPP value
	movwf	vppValue
	return


delay_us
	bcf	T1CON,TMR1ON
	bcf	PIR1,TMR1IF
	clrf	TMR1L
	clrf	TMR1H
	movf	command+1,W
	subwf	TMR1L,F
	movf	command+2,W
	subwfb	TMR1H,F
	bsf	T1CON,TMR1ON
	btfss	PIR1,TMR1IF
	goto	$-1			;wait for time over
	return

cmdBootldr
	clrf	INTCON
	bsf	FSR0,7
	goto	0

sendNAK
	movlw	NAK
	goto	txByte
sendACK
	movlw	ACK
txByte
 ifdef	SIMULATOR
	return
 endif
	banksel	BAUDCON

	btfss	TXSTA,TRMT   
	goto	txSkipBrk		;skip break if already transmitting

	clrf	temp
	incfsz	temp,F			;force a little break after previous Rx op.
	goto	$-1

	btfss	BAUDCON,RCIDL
	goto	txByte			;wait if Rx is busy

txSkipBrk
	banksel	PIR1
	btfss	PIR1,TXIF
	goto	$-1			;wait for Tx-register empty

	banksel	TXREG
	movwf	TXREG
	incf	halfDplxCnt,F
	banksel	0
	return

;--------------------- run target application
cmdRun	
	banksel	LATA
	clrf	setVP+1			;boost converter off
	bcf	flags,VPPRDY
	bsf	TRES_PORT,TRES_PIN	;apply reset (MCLR = low)
	bcf	SWVPP_PORT,SWVPP_PIN	;VPP on -> discharge cap faster
	bsf	LED_PORT,LED_PIN	;LED = status of VPP off

	banksel	0	
	movlw	high 20000		;wait 20 ms
	movwf	command+2
	call	delay_us

	bsf	SWVPP_PORT,SWVPP_PIN	;VPP off (MCLR = low)
	call	wait10us
	bcf	SWVDD_PORT,SWVDD_PIN	;VDD on
	call	wait10us
	bcf	TRES_PORT,TRES_PIN	;release reset 
	bcf	SWVPP_PORT,SWVPP_PIN	;VPP on -> pulls MCLR high
	
	banksel	TRISA
	bsf	PDAT_PORT,PDAT_PIN	;release ICSPCLK,ICSPDATA
	bsf	PCLK_PORT,PCLK_PIN

	banksel	0
	lsrf	PORTA,W			;bit 0,1 = state of ICSPCLK, ICSPDAT
	movwf	w16			;save current state

cmdRunLp
	lsrf	PORTA,W			;bit 0,1 = state of ICSPCLK, ICSPDAT
	xorwf	w16,W
	andlw	0x03			;any change of state?
	btfsc	STATUS,Z
	goto	cmdRunSkip		;no change

	xorwf	w16,F			;actualize to new state
	call	txByte			;send new state to host
cmdRunSkip
	movf	fifoRd,W
	subwf	fifoWr,W
	btfss	STATUS,Z		;any rx data in buffer?
	btfsc	STATUS,Z
	goto	cmdRunLp		;no: stay in this loop

cmdRunExit
	goto	exitPrgMode
	

;--------------------- detect target connection
;This routine detects connection of externally powered target
;and detects eventually twisted ICSP connector.
;It cannot detect connection of non-powered targets.

checkConnection
	btfsc	flags,VPPCTRL
	goto	clrConState		;VPP activation in progress

	btfss	flags,TRGCON
	goto	chkC_notconnected	;not connected (yet)

	btfss	flags,TRGXV		;connected with external VDD?
	goto	clrConState		;VDD from programmer: cannot detect disconnection here

	banksel	LATA
	bcf	VDD_PORT,VDD_PIN	
	banksel	TRISA
	bcf	VDD_PORT,VDD_PIN	;output low
	call	wait1us
	bsf	VDD_PORT,VDD_PIN	;release port
	call	wait1us
	banksel	0
	btfsc	VDD_PORT,VDD_PIN	;VDD still applied by target?
	goto	clrConState		;y: still connected

	bcf	flags,TRGCON
	bcf	flags,TRGXV		;connection lost				
	goto	clrConState			

chkC_notconnected
	movlw	conTimer
	call	check_time		;C=1 if timeout
	movf	conState,W
	btfsc	STATUS,Z
	bsf	STATUS,C		;no wait for timeout if State=0
	btfss	STATUS,C
	return				;wait for timeout...

	brw
	goto	chkConS0
	goto	chkConS1
	goto	chkConS2
clrConState
	banksel	0
	clrf	conState		;reset state
	return


chkConS0
	movlw	20
	call	set_time_fsr0

	clrf	setVP+1
	clrf	setVP
	bcf	flags,TRGXV		;default: no external target VDD detected

	banksel	LATA
	bcf	SWVPP_PORT,SWVPP_PIN	;VPP off
	bsf	TRES_PORT,TRES_PIN	;pull target MCLR low
	bsf	SWVDD_PORT,SWVDD_PIN	;target VDD = off
	bcf	VDD_PORT,VDD_PIN	;output low
	bcf	PCLK_PORT,PCLK_PIN	;ICSPCLK = low

	banksel	TRISA
	bsf	PDAT_PORT,PDAT_PIN	;ICSPDAT = input
	bcf	VDD_PORT,VDD_PIN	;VDD sensing: output low
	banksel	0
	movlw	0xFF
	movwf	conVar+1		;init to high value
incConState
	banksel	0
	incf	conState,F
	return

;state 1: outputs low or off, wait until VP voltage stabilizes (capacitor discharged) 
chkConS1
	movlw	5
	call	set_time_fsr0
	
	movf	conVar,W		;ad_res - previous value:
	bcf	INTCON,GIE
	subwf	ad_res,W		;is voltage still dropping?
	movf	conVar+1,W
	subwfb	ad_res+1,W		;(C=0 if falling)
	bsf	INTCON,GIE
	mov16	ad_res,conVar		;store actual value
	btfss	STATUS,C		
	return				;voltage still dropping

	btfss	PDAT_PORT,PDAT_PIN
	goto	chkConS1Datlow		;as normally expected: ICSPDAT is low by pull down

;ICSPDAT Pin driven high from target: check if connector is probably twisted
	banksel	TRISA
	bsf	VDD_PORT,VDD_PIN	;-> input floating
	call	wait1us
	banksel	0
	btfss	VDD_PORT,VDD_PIN
	goto	con_twisted		;ICSPDAT high, but VDD low: must be wrong!

;ICSPDAT and VDD are both high, maybe target has strong pullup on ICSPDAT.
;in this case, VDD level must be higher than ICSPDAT  
	movlw	2			;AD channel: ICSPDAT
	call	adConvert	
	mov16	w16,conVar	
	movlw	7			;AD channel: Target VDD
	call	adConvert

	banksel	0			;compare levels on pins:
	movf	conVar,W		;level(VDD) - level(ICSPDAT)
	subwf	w16,W
	movf	conVar+1,W		
	subwfb	w16+1,W
	btfss	STATUS,C
	goto	con_twisted		;level (ICSPDAT) higher than level(VDD)

conTargetVddExtern
	bcf	flags,CONERR		;assumed that plug is not twisted

	btfsc	flags,TRGXV		;try again after time for stabilizing connection
	bsf	flags,TRGCON		;assume target connected
	bsf	flags,TRGXV		;it's external target VDD
	incf	conTimer+1,F		;1/4 sec. for stabilization of contacts
	 
	btfsc	flags,TRGCON		;detection complete if flag = "connected"
	clrf	conState
;	banksel	TRISA
;	bcf	VDD_PORT,VDD_PIN	;output (low)
	banksel	0
	return
	
con_twisted
	bsf	flags,CONERR
	goto	clrConState
		
chkConS1Datlow
	banksel	TRISA			;check if VDD is externally pulled high:
	bcf	INTCON,GIE
	bsf	VDD_PORT,VDD_PIN	;change (out low) to (input floating)
	banksel	0
	nop
	nop
	movf	VDD_PORT,W
	bsf	INTCON,GIE

	btfsc	WREG,VDD_PIN
	goto 	conTargetVddExtern	;target VDD supplied external

	banksel	LATA
	bsf	SWVPP_PORT,SWVPP_PIN	;VPP on (voltage is less than VDD)
	bcf	TRES_PORT,TRES_PIN	;release MCLR

	banksel	TRISA
	bcf	PDAT_PORT,PDAT_PIN	;drive ICSPDAT low
	goto	incConState

;VPP is turned on. If connection is wrong, there will be a large voltage drop by
;current through port input diodes.
chkConS2
	bcf	INTCON,GIE
	banksel	LATA
	bcf	SWVPP_PORT,SWVPP_PIN	;VPP off 
	banksel	0
	movf	ad_res,W		;previous value - ad_res
	subwf	conVar,F		;-> conVar (voltage drop)
	movf	ad_res+1,W
	bsf	INTCON,GIE
	subwfb	conVar+1,F		
	btfss	STATUS,C
	goto	chkConS2_lowI		;no voltage drop at all

	movlw	low IVPP_LIMIT
	subwf	conVar,W
	movlw	high IVPP_LIMIT
	subwfb	conVar+1,W
	btfsc	STATUS,C
	goto	con_twisted		;voltage drop >= IVPP_LIMIT

chkConS2_lowI				;detected VPP current is low:
	bcf	flags,CONERR		;no connection or proper connection without external VDD
	goto	clrConState


;--------------------- AD conversion (channel W), result -> w16
adConvert	
	aslf	WREG,W
	aslf	WREG,W			;ch -> channel select bits
	iorlw	1			;AD on
	banksel	ADCON0
adBusy
	bsf	INTCON,GIE
	btfsc	ADCON0,GO		
	goto	adBusy			;currently an AD conversion is going on!	

	bcf	INTCON,GIE
	btfsc	ADCON0,GO		
	goto	adBusy			; AD conversion just started by ISR

	bsf	flags,ADLOCK		;lock AD converter for ISR
	banksel	PIE1
	bcf	PIE1,ADIE		;disable AD interrupt
	bsf	INTCON,GIE

	banksel	ADCON0
	movwf	ADCON0			;select channel
	movlw	b'11100000'		;right just., 0.5MHz, PREF=VDD
	movwf	ADCON1
	
	call	wait8us			;wait: aquisition time 

	bsf	ADCON0,GO		;start conversion

	btfsc	ADCON0,GO
	goto	$-1			;wait for conversion complete

	mov16	ADRESL,w16		;store result

;restore standard AD-setting:
	movlw	1
	movwf	ADCON0			;select input: AN0, enable ADC 
	movlw	b'11100011'		;right just., 0.5MHz, PREF=VFR
	movwf	ADCON1

	call	wait6us			;wait: aquisition time 

	banksel	PIR1
	bcf	PIR1,ADIF

	banksel	PIE1
	bsf	PIE1,ADIE		;enable AD interrupt
	bcf	flags,ADLOCK
	banksel	0
	return
		

		
	

;--------------------- Set Vpp to value W (* 0.1 V)
SetVpp
	banksel 0
	clrf	r16_c+1			;rc = requested voltage
	movwf	r16_c
	clrf	r16_b+1			;rb = 100 (-> 10V)
	movlw	100
	movwf	r16_b
	movlw	high 896		;calibrated AD value @10V
	movwf	r16_a+1			; -> ra
	movlw	low 896
	movwf	r16_a
	call	proportion		;rx = (ra/rb)*rc

	movf	r16_x,W
	bcf	INTCON,GIE
	movwf	setVP
	movf	r16_x+1,W
	movwf	setVP+1
	bsf	INTCON,GIE
	return

;----------------------------- enter programming mode
enterPrgMode
	btfsc	flags,CONERR
	goto	sendNAK			;abort if connection is wrong

	bsf	flags,TRGCON		;assume that target is now connected

	banksel	LATA
	bcf	SWVPP_PORT,SWVPP_PIN
	bcf	VDD_PORT,VDD_PIN	;VPP low	
	bsf	TRES_PORT,TRES_PIN	;for sure: apply target reset	
	bcf	PDAT_PORT,PDAT_PIN	;drive both ICSP-lines low
	bcf	PCLK_PORT,PCLK_PIN
	banksel	TRISA
	bcf	PDAT_PORT,PDAT_PIN
	bcf	PCLK_PORT,PCLK_PIN

	btfss	flags,TRGXV		;target VDD supplied externally?
	bcf	VDD_PORT,VDD_PIN	;no: keep VDD still pulled low 

	banksel	0
	bsf	flags,VPPCTRL
	movf	vppValue,W
	goto	SetVpp

PrgModeWhenReady
	btfss	flags,VPPRDY
	return				;wait for VPP reaching set level

	bcf	flags,VPPCTRL
	bcf	TRES_PORT,TRES_PIN	;release target reset	
	call	wait4us
	bsf	SWVPP_PORT,SWVPP_PIN	;rise VPP
	bsf	flags,VPPACTV
	
	call	sendACK

	btfss	flags,TRGXV		;target VDD supplied externally?
	bcf	SWVDD_PORT,SWVDD_PIN	;no: enable VDD
	banksel	TRISA
	bcf	VDD_PORT,VDD_PIN	;release VDD pull down  
	banksel	0

	movlw	25			;wait >= 250s...
tenth_wt
	call	wait10us
	decfsz	WREG,W
	goto	tenth_wt

	return

;----------------------------- exit programming mode
exitPrgMode
	banksel	LATA
	bsf	SWVDD_PORT,SWVDD_PIN	;VDD off
	bcf	flags,VPPACTV
	bcf	flags,VPPCTRL
	bcf	VDD_PORT,VDD_PIN	;pull down VDD line
	bcf	PDAT_PORT,PDAT_PIN	;drive both ICSP-lines low
	bcf	PCLK_PORT,PCLK_PIN
	banksel	TRISA
	bcf	VDD_PORT,VDD_PIN
	bcf	PDAT_PORT,PDAT_PIN
	bcf	PCLK_PORT,PCLK_PIN
	banksel	0
	call	wait10us

	bcf	SWVPP_PORT,SWVPP_PIN	;VPP off
	clrf	setVP
	clrf	setVP+1
	bsf	TRES_PORT,TRES_PIN	;apply target reset
	bcf	flags,VPPRDY

	btfss	flags,TRGXV		;target VDD supplied externally?
	bcf	flags,TRGCON		;no: assume target gets disconnected
	return	

;----------------------------- send an "increment address" command
incAddr
	movlw	6
	goto	sendCmdW

;----------------------------- send command (command)
sendCommand
	btfss	flags,TRGCON
	return				;(no target connected)

	movf	command,W
sendCmdW
	movwf	temp

	banksel	LATA
	bcf	PDAT_PORT,PDAT_PIN	;ICSPDAT = low
	banksel	TRISA
	bcf	PDAT_PORT,PDAT_PIN
	banksel	LATA
	
	movlw	6			;count 6 clock pulses
sendCmdLp
	rrf	temp,F		;LSB -> C
	btfsc	STATUS,C
	bsf	PDAT_PORT,PDAT_PIN	;C-> ISCPDAT
	bsf	PCLK_PORT,PCLK_PIN	;clock pulse...
	btfss	STATUS,C
	bcf	PDAT_PORT,PDAT_PIN
	nop
	bcf	PCLK_PORT,PCLK_PIN
	decfsz	WREG,F
	goto	sendCmdLp		;next bit

releaseICSPDAT
	banksel	TRISA
	bsf	PDAT_PORT,PDAT_PIN	;release ICSPDAT
	banksel	0
	goto	wait1us
	

;----------------------------- send data (w16, 14 bits)
sendICSPDAT
	btfss	flags,TRGCON
	return				;(no target connected)

	movlw	16			;send with 16 clocks
	movwf	temp

	bcf	STATUS,C		;start with sending a zero as startbit

	banksel	TRISA
	bcf	PDAT_PORT,PDAT_PIN	;drive ISCPDAT
	banksel	LATA	
sendDatLp
	btfsc	STATUS,C
	bsf	PDAT_PORT,PDAT_PIN	;C-> ISCPDAT
	bsf	PCLK_PORT,PCLK_PIN	;clock pulse
	btfss	STATUS,C
	bcf	PDAT_PORT,PDAT_PIN
	lsrf	w16+1,F
	rrf	w16,F		;next bit -> C
	bcf	PCLK_PORT,PCLK_PIN
	decfsz	temp,F
	goto	sendDatLp
	banksel	0
	return

;----------------------------- receive data (w16, 14 bits)
getICSPDAT
	movlw	16			;count 16 clocks
	movwf	temp
	clrf	w16+1

getDatLp
	bsf	PCLK_PORT,PCLK_PIN	;clock pulse
	lsrf	w16+1,F			
	rrf	w16,F	
	btfsc	PDAT_PORT,PDAT_PIN	;bit14(result) = ISCPDAT
	bsf	w16+1,6
	bcf	PCLK_PORT,PCLK_PIN
	decfsz	temp,F
	goto	getDatLp

	bcf	w16+1,6			;16th bit is stopbit
	return		
	

#include timer.asm
#include math.asm

	END